/*
 * Copyright (C) 2015 Nu Development Team
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.nubits.nubot.trading.wrappers;


import com.nubits.nubot.bot.Global;
import com.nubits.nubot.exchanges.ExchangeFacade;
import com.nubits.nubot.global.Constant;
import com.nubits.nubot.global.Settings;
import com.nubits.nubot.models.*;
import com.nubits.nubot.models.Currency;
import com.nubits.nubot.trading.*;
import com.nubits.nubot.trading.keys.ApiKeys;
import com.nubits.nubot.utils.TimeUtils;
import com.nubits.nubot.utils.Utils;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;
import org.json.simple.parser.ParseException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.net.ssl.HttpsURLConnection;
import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.NoRouteToHostException;
import java.net.URL;
import java.net.UnknownHostException;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.*;

public class SleightWrapper implements TradeInterface {

    private static final Logger LOG = LoggerFactory.getLogger(SleightWrapper.class.getName());
    private ApiKeys keys;
    protected SleightService service;
    public final String EXCHANGE_NAME = ExchangeFacade.SLEIGHT;

    private final String SIGN_HASH_FUNCTION = "HmacSHA256";
    private final String ENCODING = "UTF-8";
    //handle concurrent requests
    private boolean isBusy;

    //Entry points
    private final String API_BASE_URL = "https://sleight.herokuapp.com/";
    //private final String API_BASE_URL = "https://local.wazoku.com/";
    private String checkConnectionUrl = API_BASE_URL;
    private final String API_GET_BALANCES = "get_balances";
    private final String API_GET_ORDERS = "get_orders";
    private final String API_GET_TRADES = "get_trades";
    private final String API_SELL = "ask";
    private final String API_BUY = "bid";
    private final String API_CANCEL_ORDER = "cancel_order";
    private final String API_TICKER = "get_ticker";
    private final String API_ORDERBOOK = "get_order_book";
    private final String API_PLACE_ORDER = "place_order";


    //Errors
    private ErrorManager errors = new ErrorManager();
    private final String TOKEN_BAD_RETURN = "No Connection With Exchange";

    private long nonceCount = new Long(System.currentTimeMillis() / 100000).longValue();
    private boolean fixNonce = false;

    public SleightWrapper(ApiKeys keys) {
        this.keys = keys;
        service = new SleightService(keys);
        setupErrors();
        this.setFree();
    }

    private void setupErrors() {
        errors.setExchangeName(this.EXCHANGE_NAME);
    }

    private ApiResponse getQuery(String url, String method, HashMap<String, String> query_args, boolean needAuth, boolean isGet) {
        //LOG.info("calling " + url + " with " + method + " and " + query_args);
        ApiResponse apiResponse = new ApiResponse();
        String queryResult = query(url, method, query_args, needAuth, isGet);
        if (queryResult == null) {
            apiResponse.setError(errors.nullReturnError);
            return apiResponse;
        }
        if (queryResult.equals(TOKEN_BAD_RETURN)) {
            apiResponse.setError(errors.noConnectionError);
            return apiResponse;
        }

        JSONParser parser = new JSONParser();
        try {
            JSONObject httpAnswerJson = (JSONObject) (parser.parse(queryResult));
            Boolean success = (Boolean) httpAnswerJson.get("success");
            if (success.equals(false)) {
                String errorMessage = httpAnswerJson.get("message").toString();
                ApiError apiErr = errors.apiReturnError;
                apiErr.setDescription(errorMessage);
                LOG.debug("Sleight API returned an error: " + errorMessage);
                apiResponse.setError(apiErr);
            } else {
                apiResponse.setResponseObject(httpAnswerJson.get("message"));
            }
        } catch (ParseException ex) {
            LOG.error("httpresponse: " + queryResult + " \n" + ex.toString());
            apiResponse.setError(errors.parseError);
            return apiResponse;
        }
        return apiResponse;
    }

    @Override
    public ApiResponse getAvailableBalances(CurrencyPair pair) {
        return getBalanceImpl(pair, null);
    }

    @Override
    public ApiResponse getAvailableBalance(Currency currency) {
        return getBalanceImpl(null, currency);
    }

    @Override
    public ApiResponse placeOrders(OrderBatch batch, CurrencyPair pair) {
        return TradeUtils.placeMultipleOrdersSequentiallyImplementation(batch, pair, TradeUtils.INTERVAL_FAST, this);
    }

    private ApiResponse getBalanceImpl(CurrencyPair pair, Currency currency) {
        LOG.trace("get balance");

        //Swap the pair for the request
        ApiResponse apiResponse = new ApiResponse();

        String url = API_BASE_URL;
        String method = API_GET_BALANCES;
        HashMap<String, String> query_args = new HashMap<>();
        boolean isGet = false;

        LOG.trace("get from " + url);
        LOG.trace("method " + method);
        ApiResponse response = getQuery(url, method, query_args, true, isGet);

        LOG.trace("response " + response);
        if (!response.isPositive()) {
            return response;
        }

        JSONObject httpAnswerJson = (JSONObject) response.getResponseObject();
        JSONArray wallets = (JSONArray) httpAnswerJson.get("balances");
        LOG.trace("balance answer " + wallets);

        if (currency != null) {
            //looking for a specific currency
            Boolean found = false;
            Amount balance = new Amount(0, currency);
            for (Iterator<JSONObject> wallet = wallets.iterator(); wallet.hasNext(); ) {
                JSONObject thisWallet = wallet.next();
                if (thisWallet.get("currency_code").equals(currency.getCode().toUpperCase())) {
                    balance.setQuantity(Utils.getDouble(thisWallet.get("amount")));
                    found = true;
                }
            }
            apiResponse.setResponseObject(balance);
            if (!found) {
                String errorMessage = "Cannot find a balance for currency " + currency.getCode().toUpperCase();
                ApiError apiErr = errors.apiReturnError;
                apiErr.setDescription(errorMessage);
                apiResponse.setError(apiErr);
            }
        } else {
            //get all balances for the pair
            boolean foundNBTavail = false;
            boolean foundPEGavail = false;
            Amount NBTAvail = new Amount(0, pair.getOrderCurrency());
            Amount PEGAvail = new Amount(0, pair.getPaymentCurrency());

            Amount PEGonOrder = new Amount(0, pair.getPaymentCurrency());
            Amount NBTonOrder = new Amount(0, pair.getOrderCurrency());

            String NBTcode = pair.getOrderCurrency().getCode().toUpperCase();
            String PEGcode = pair.getPaymentCurrency().getCode().toUpperCase();

            for (Iterator<JSONObject> wallet = wallets.iterator(); wallet.hasNext(); ) {
                JSONObject thisWallet = wallet.next();
                if (thisWallet.get("currency_code").equals(NBTcode)) {
                    NBTAvail.setQuantity(Utils.getDouble(thisWallet.get("amount")));
                    foundNBTavail = true;
                }
                if (thisWallet.get("currency_code").equals(PEGcode)) {
                    PEGAvail.setQuantity(Utils.getDouble(thisWallet.get("amount")));
                    foundPEGavail = true;
                }
            }

            PairBalance balance = new PairBalance(PEGAvail, NBTAvail, PEGonOrder, NBTonOrder);
            apiResponse.setResponseObject(balance);
            if (!foundNBTavail || !foundPEGavail) {
                LOG.warn("Cannot find a balance for currency with code "
                        + "" + NBTcode + " or " + PEGcode + " in your balance. "
                        + "NuBot assumes that balance is 0");
            }
        }

        return apiResponse;
    }

    @Override
    public ApiResponse getLastPrice(CurrencyPair pair) {
        Ticker ticker = new Ticker();
        ApiResponse apiResponse = new ApiResponse();

        double last = -1;
        double ask = -1;
        double bid = -1;


        //https://poloniex.com/public?command=returnTicker
        String url = API_BASE_URL;
        String method = API_TICKER + '/' + pair.getOrderCurrency() + '/' + pair.getPaymentCurrency();
        HashMap<String, String> query_args = new HashMap<>();
        boolean isGet = true;

        LOG.trace("get from " + url);
        LOG.trace("method " + method);
        ApiResponse response = getQuery(url, method, query_args, false, isGet);

        LOG.trace("response " + response);
        if (!response.isPositive()) {
            return response;
        } else {
            /* Sample response
                'ticker': [
                    {
                        "last_trade_price": "0.00000035",
                        "lowest_ask_price": "0.00000036",
                        "highest_bid_price": "0.00000035",
                        "base_volume": "0.15543370",
                        "relative_volume": "512450.46072496",
                        "24_hour_high": "0.00000036",
                        "24_hour_low": "0.00000021"
                    }
                ]
            */
            JSONObject httpAnswerJson = (JSONObject) response.getResponseObject();
            JSONArray ticker_data = (JSONArray) httpAnswerJson.get("ticker");

            String key = pair.toStringSepSpecial("/").toLowerCase();
            for (Iterator<JSONObject> data = ticker_data.iterator(); data.hasNext(); ) {
                JSONObject thisData = data.next();

                last = Utils.getDouble(thisData.get("last_trade_price"));
                bid = Utils.getDouble(thisData.get("highest_bid_price"));
                ask = Utils.getDouble(thisData.get("lowest_ask_price"));

                ticker.setAsk(ask);
                ticker.setBid(bid);
                ticker.setLast(last);
            }
            apiResponse.setResponseObject(ticker);
            return apiResponse;
        }
    }


    @Override
    public ApiResponse sell(CurrencyPair pair, double amount, double rate) {
        return enterOrder(Constant.SELL, pair, amount, rate);
    }

    @Override
    public ApiResponse buy(CurrencyPair pair, double amount, double rate) {
        return enterOrder(Constant.BUY, pair, amount, rate);
    }

    private ApiResponse enterOrder(String type, CurrencyPair pair, double amount, double rate) {
        ApiResponse apiResponse = new ApiResponse();
        boolean isGet = false;
        String url = API_BASE_URL;
        String method = API_PLACE_ORDER;

        HashMap<String, String> query_args = new HashMap<>();

        /*Params
         */
        if (type.equals(Constant.SELL)) {
            query_args.put("order_type", API_SELL);
            query_args.put("amount", Double.toString(amount));
        } else {
            query_args.put("order_type", API_BUY);
            query_args.put("amount", Double.toString(amount));
        }
        pair = CurrencyPair.swap(pair);
        query_args.put("pair", pair.toStringSepSpecial("/").toLowerCase());
        query_args.put("price", Double.toString(rate));

        ApiResponse response = getQuery(url, method, query_args, true, isGet);
        if (response.isPositive()) {
            JSONObject httpAnswerJson = (JSONObject) response.getResponseObject();
            JSONArray orders = (JSONArray) httpAnswerJson.get("orders");
            Long order_id_long = (Long) httpAnswerJson.get("order_id");
            String order_id = Long.toString(order_id_long);
            apiResponse.setResponseObject(order_id);
        } else {
            apiResponse = response;
        }

        return apiResponse;
    }

    @Override
    public ApiResponse getActiveOrders() {
        return getOrdersImpl(null);
    }

    @Override
    public ApiResponse getActiveOrders(CurrencyPair pair) {
        return getOrdersImpl(pair);
    }

    private ApiResponse getOrdersImpl(CurrencyPair pair) {
        ApiResponse apiResponse = new ApiResponse();
        ArrayList<Order> orderList = new ArrayList<Order>();
        boolean isGet = false;
        String url = API_BASE_URL;
        String method = API_GET_ORDERS;
        HashMap<String, String> query_args = new HashMap<>();

        ApiResponse response = getQuery(url, method, query_args, true, isGet);
        if (response.isPositive()) {
            JSONObject httpAnswerJson = (JSONObject) response.getResponseObject();
            JSONArray orders = (JSONArray) httpAnswerJson.get("orders");
            for (Iterator<JSONObject> order = orders.iterator(); order.hasNext(); ) {
                JSONObject orderObject = order.next();
                CurrencyPair cp = CurrencyPair.getCurrencyPairFromString(orderObject.get("pair").toString().toUpperCase(), "/");
                Order tempOrder = parseOrder(orderObject, cp);
                if (tempOrder.isCompleted()){
                    continue;
                }
                if (pair == null) {
                    orderList.add(tempOrder);
                } else {
                    if (pair.equals(CurrencyPair.swap(cp))) {
                        orderList.add(tempOrder);
                    }
                }
            }
            apiResponse.setResponseObject(orderList);
        } else {
            apiResponse = response;
        }

        return apiResponse;
    }

    @Override
    public ApiResponse getOrderDetail(String orderID) {
        ApiResponse apiResp = new ApiResponse();
        Order order = null;

        ApiResponse listApiResp = getActiveOrders();
        if (listApiResp.isPositive()) {
            ArrayList<Order> orderList = (ArrayList<Order>) listApiResp.getResponseObject();
            boolean found = false;
            for (int i = 0; i < orderList.size(); i++) {
                Order tempOrder = orderList.get(i);
                if (orderID.equals(tempOrder.getId())) {
                    found = true;
                    apiResp.setResponseObject(tempOrder);
                    return apiResp;
                }
            }
            if (!found) {
                ApiError apiErr = errors.apiReturnError;
                apiErr.setDescription("Cannot find the order with id " + orderID);
                apiResp.setError(apiErr);
                return apiResp;

            }
        } else {
            return listApiResp;
        }

        return apiResp;
    }

    @Override
    public ApiResponse cancelOrder(String orderID, CurrencyPair pair) {
        ApiResponse apiResponse = new ApiResponse();
        String url = API_BASE_URL;
        boolean isGet = false;
        String method = API_CANCEL_ORDER;
        HashMap<String, String> query_args = new HashMap<>();

        query_args.put("order_id", orderID);

        ApiResponse response = getQuery(url, method, query_args, true, isGet);
        if (response.isPositive()) {
            apiResponse.setResponseObject(true);
        } else {
            apiResponse = response;
        }

        return apiResponse;
    }

    @Override
    public ApiResponse getTxFee() {
        return new ApiResponse(true, Global.options.getTxFee(), null);
    }

    @Override
    public ApiResponse getTxFee(CurrencyPair pair) {
        LOG.debug("Poloniex uses global TX fee, currency pair not supprted. \n" + "now calling getTxFee()");
        return getTxFee();
    }

    @Override
    public ApiResponse isOrderActive(String id) {
        ApiResponse existResponse = new ApiResponse();

        ApiResponse orderDetailResponse = getOrderDetail(id);
        if (orderDetailResponse.isPositive()) {
            Order order = (Order) orderDetailResponse.getResponseObject();
            existResponse.setResponseObject(true);
        } else {
            ApiError err = orderDetailResponse.getError();
            if (err.getDescription().contains("Cannot find the order")) {
                existResponse.setResponseObject(false);

            } else {
                existResponse.setError(err);
            }
        }
        return existResponse;
    }

    @Override
    public ApiResponse clearOrders(CurrencyPair pair) {
        //Since there is no API entry point for that, this call will iterate over active orders
        ApiResponse toReturn = new ApiResponse();
        boolean ok = true;

        ApiResponse activeOrdersResponse = getActiveOrders();
        if (activeOrdersResponse.isPositive()) {
            ArrayList<Order> orderList = (ArrayList<Order>) activeOrdersResponse.getResponseObject();
            for (int i = 0; i < orderList.size(); i++) {
                Order tempOrder = orderList.get(i);
                ApiResponse deleteOrderResponse = cancelOrder(tempOrder.getId(), pair);
                if (deleteOrderResponse.isPositive()) {
                    boolean deleted = (boolean) deleteOrderResponse.getResponseObject();

                    if (deleted) {
                        LOG.warn("Order " + tempOrder.getId() + " deleted succesfully");
                    } else {
                        LOG.error("Could not delete order " + tempOrder.getId() + "");
                        ok = false;
                    }

                } else {
                    LOG.error(deleteOrderResponse.getError().toString());
                }
                try {
                    Thread.sleep(500);
                } catch (InterruptedException ex) {
                    LOG.error(ex.toString());
                }

            }
            toReturn.setResponseObject(ok);
        } else {
            LOG.error(activeOrdersResponse.getError().toString());
            toReturn.setError(activeOrdersResponse.getError());
            return toReturn;
        }

        return toReturn;
    }

    @Override
    public ApiError getErrorByCode(int code) {
        return null;
    }

    @Override
    public String getUrlConnectionCheck() {
        return checkConnectionUrl;
    }

    @Override
    public ApiResponse placeOrdersParallel(OrderBatch batch, CurrencyPair pair, ArrayList<ApiKeys> keys) {
        return TradeUtils.placeMultipleOrdersParallelImplementation(batch, pair, keys, this);
    }

    @Override
    public String query(String base, String method, AbstractMap<String, String> args, boolean needAuth, boolean isGet) {
        String queryResult = TOKEN_BAD_RETURN; //Will return this string in case it fails
        if (this.isFree()) {
            this.setBusy();
            queryResult = service.executeQuery(base, method, args, needAuth, isGet);
            this.setFree();
        } else {
            //Another thread is probably executing a query. Init the retry procedure
            long sleeptime = Settings.RETRY_SLEEP_INCREMENT * 1;
            int counter = 0;
            long startTimeStamp = System.currentTimeMillis();
            LOG.debug(method + " blocked, another call is being processed ");
            boolean exit = false;
            do {
                counter++;
                sleeptime = counter * Settings.RETRY_SLEEP_INCREMENT; //Increase sleep time
                sleeptime += (int) (Math.random() * 200) - 100;// Add +- 100 ms random to facilitate competition
                LOG.debug("Retrying for the " + counter + " time. Sleep for " + sleeptime + "; Method=" + method);
                try {
                    Thread.sleep(sleeptime);
                } catch (InterruptedException e) {
                    LOG.error(e.toString());
                }

                //Try executing the call
                if (this.isFree()) {
                    LOG.debug("Finally the exchange is free, executing query after " + counter + " attempt. Method=" + method);
                    this.setBusy();
                    queryResult = service.executeQuery(base, method, args, needAuth, isGet);
                    this.setFree();
                    break; //Exit loop
                } else {
                    LOG.debug("Exchange still busy : " + counter + " .Will retry soon; Method=" + method);
                    exit = false;
                }
                if (System.currentTimeMillis() - startTimeStamp >= Settings.TIMEOUT_QUERY_RETRY) {
                    exit = true;
                    LOG.error("Method=" + method + " failed too many times and timed out. attempts = " + counter);
                }
            } while (!exit);
        }

        return queryResult;
    }

    @Override
    public SignedRequest getOpenOrdersRequest(CurrencyPair pair) {
        //Create the vocabulary
        LinkedHashMap<String, String> query_args = new LinkedHashMap<>();

        String pairString = "all";
        if (pair != null) {
            pair = CurrencyPair.swap(pair);
            pairString = pair.toStringSep().toUpperCase();
        }

        query_args.put("nonce", createNonce());
        query_args.put("command", API_GET_ORDERS);
        query_args.put("currencyPair", pairString);

        //Create the signature
        String post_data = TradeUtils.buildQueryString(query_args, ENCODING);
        String signature = TradeUtils.signRequest(keys.getPrivateKey(), post_data, SIGN_HASH_FUNCTION, ENCODING);

        return new SignedRequest(query_args, signature);
    }


    @Override
    public void setKeys(ApiKeys keys) {
        this.keys = keys;
    }

    @Override
    public void setApiBaseUrl(String apiBaseUrl) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public ApiResponse getLastTrades(CurrencyPair pair) {
        return getTradesImpl(pair, 0);
    }

    @Override
    public ApiResponse getLastTrades(CurrencyPair pair, long startTime) {
        startTime = Math.round(startTime / 1000);
        return getTradesImpl(pair, startTime);
    }

    private ApiResponse getTradesImpl(CurrencyPair pair, long startTime) {
        ApiResponse apiResponse = new ApiResponse();
        ArrayList<Trade> tradeList = new ArrayList<Trade>();
        boolean isGet = false;
        String url = API_BASE_URL;
        String method = API_GET_TRADES;
        HashMap<String, String> query_args = new HashMap<>();

        String startDateArg;
        if (startTime == 0) {
            long now = System.currentTimeMillis();
            long yesterday = Math.round((now - TimeUtils.getOneDayInMillis()) / 1000);
            startDateArg = Long.toString(yesterday); //24hours
        } else {
            startDateArg = Long.toString(startTime);
        }

        pair = CurrencyPair.swap(pair);
        query_args.put("currencyPair", pair.toStringSep().toUpperCase());
        query_args.put("start", startDateArg);

        ApiResponse response = getQuery(url, method, query_args, true, isGet);
        if (response.isPositive()) {
            JSONArray httpAnswerJson = (JSONArray) response.getResponseObject();
            for (int i = 0; i < httpAnswerJson.size(); i++) {
                JSONObject tradesObject = (JSONObject) httpAnswerJson.get(i);
                Trade tempTrade = parseTrade(tradesObject, pair);
                tradeList.add(tempTrade);
            }
            apiResponse.setResponseObject(tradeList);
        } else {
            apiResponse = response;
        }

        return apiResponse;
    }

    private Order parseOrder(JSONObject orderObject, CurrencyPair pair) {
        /* {"amount":0.0,"price":8.0,"original_amount":25.0,"id":13,"state":"complete","order_type":"bid","pair":"btc\/usnbt"} */
        Order order = new Order();

        order.setType(((String) orderObject.get("order_type")).toUpperCase().equals("ASK") ? Constant.SELL : Constant.BUY);
        order.setId(orderObject.get("id").toString());
        order.setAmount(new Amount(Utils.getDouble(orderObject.get("amount")), pair.getPaymentCurrency()));
        order.setPrice(new Amount(Utils.getDouble(orderObject.get("price")), pair.getOrderCurrency()));
        order.setCompleted(orderObject.get("state").toString().equals("complete") || orderObject.get("state").toString().equals("cancelled"));
        order.setPair(pair);
        order.setInsertedDate(new Date()); //Not provided

        return order;

    }

    private Trade parseTrade(JSONObject tradeObj, CurrencyPair pair) {
        /* {"date":"2014-02-19 04:55:44","rate":"0.0015","amount":"100","fee":"0.02","total":"0.15","orderNumber":"3048903","type":"sell"}*/
        Trade trade = new Trade();
        trade.setOrder_id((String) tradeObj.get("orderNumber"));

        trade.setExchangeName(ExchangeFacade.POLONIEX);
        trade.setPair(pair);

        trade.setType(((String) tradeObj.get("type")).toUpperCase());
        trade.setAmount(new Amount(Utils.getDouble(tradeObj.get("amount")), pair.getPaymentCurrency()));
        trade.setPrice(new Amount(Utils.getDouble(tradeObj.get("rate")), pair.getOrderCurrency()));
        double feePercent = Utils.getDouble(tradeObj.get("fee"));
        double feePaid = Utils.round(feePercent * trade.getAmount().getQuantity());
        trade.setFee(new Amount(feePaid, pair.getPaymentCurrency())); //TODO

        String date = (String) tradeObj.get("date");
        trade.setDate(parseDate(date));

        return trade;
    }

    private Date parseDate(String dateStr) {
        Date toRet = null;
        //Parse the date
        //Sample 2014-02-19 04:55:44

        String datePattern = "yyyy-MM-dd HH:mm:ss";
        DateFormat df = new SimpleDateFormat(datePattern, Locale.ENGLISH);
        try {
            toRet = df.parse(dateStr);
        } catch (java.text.ParseException ex) {
            LOG.error(ex.toString());
            toRet = new Date();
        }
        return toRet;
    }

    @Override
    public ApiResponse getOrderBook(CurrencyPair pair) {
        ApiResponse apiResponse = new ApiResponse();
        ArrayList<Order> orderList = new ArrayList<Order>();

        //https://poloniex.com/public?command=returnTicker
        String url = "https://poloniex.com/public?command=" + API_ORDERBOOK;
        String method = "";
        HashMap<String, String> query_args = new HashMap<>();
        boolean isGet = true;

        pair = CurrencyPair.swap(pair);
        String pairString = pair.toStringSep().toUpperCase();


        url += "&currencyPair=" + pairString + "&depth=1500";

        ApiResponse response = getQuery(url, method, query_args, false, isGet);
        if (response.isPositive()) {
            JSONObject httpAnswerJson = (JSONObject) response.getResponseObject();
            ArrayList<Order> orderBook = new ArrayList<>();
            JSONArray asks = (JSONArray) httpAnswerJson.get("asks");
            JSONArray bids = (JSONArray) httpAnswerJson.get("bids");
            for (Iterator<JSONArray> order = asks.iterator(); order.hasNext(); ) {
                orderBook.add(parseOrderSimple(order.next(), Constant.SELL, pair));
            }
            for (Iterator<JSONArray> order = bids.iterator(); order.hasNext(); ) {
                orderBook.add(parseOrderSimple(order.next(), Constant.BUY, pair));
            }
            apiResponse.setResponseObject(new OrderBook(this.EXCHANGE_NAME, pair, orderBook));
        } else {
            apiResponse = response;
        }

        return apiResponse;


    }

    private Order parseOrderSimple(JSONArray in, String type, CurrencyPair pair) {
        Order out = new Order();
        pair = pair.swap(pair); //Swap for poloniex
        out.setType(type);
        out.setPair(pair);

        Amount amount = new Amount(Utils.getDouble(in.get(1)), pair.getOrderCurrency());
        out.setAmount(amount);
        Amount price = new Amount(Utils.getDouble(in.get(0)), pair.getPaymentCurrency());
        out.setPrice(price);

        out.setCompleted(false);

        return out;
    }

    @Override
    public boolean isBusy() {
        return isBusy;
    }

    @Override
    public boolean isFree() {
        return !isBusy;
    }

    @Override
    public void setBusy() {
        this.isBusy = true;
    }

    @Override
    public void setFree() {
        this.isBusy = false;
    }


    private String createNonce() {

        //potential FIX: add some time to the nonce, since time sync has issues
        //long fixtime = 1000;
        if (!fixNonce)
            return "" + System.currentTimeMillis();
        else {
            nonceCount++;
            LOG.trace("nonce used " + nonceCount);
            return "" + nonceCount;
        }
    }


    private class SleightService implements ServiceInterface {

        protected ApiKeys keys;

        private SleightService(ApiKeys keys) {
            this.keys = keys;
        }


        @Override
        public String executeQuery(String base, String method, AbstractMap<String, String> args, boolean needAuth, boolean isGet) {
            String answer = "";
            String signature = "";
            String post_data = "";
            String url = base + method;
            boolean httpError = false;
            HttpsURLConnection connection = null;
            URL queryUrl = null;

            try {
                // build URL

                queryUrl = new URL(url);
                LOG.trace("Query " + queryUrl);
                connection = (HttpsURLConnection) queryUrl.openConnection();
                if (isGet){
                    connection.setRequestMethod("GET");
                } else {
                    connection.setRequestMethod("POST");
                }
                LOG.trace("connection " + connection);
            } catch (IOException e) {
                LOG.error("can't connect to " + queryUrl);
            }


            try {
                // add nonce and build arg list
                if (needAuth) {
                    String nonce = createNonce();
                    LOG.trace("nonce used " + nonce);
                    args.put("nonce", nonce);
                    args.put("api_key", this.keys.getApiKey());
                    String nonce_hash = TradeUtils.signRequest(keys.getPrivateKey(), nonce, SIGN_HASH_FUNCTION, ENCODING);
                    args.put("nonce_hash", nonce_hash);
                }

                post_data = TradeUtils.buildQueryString(args, ENCODING);

                // create and setup a HTTP connection

                connection.setRequestProperty("Content-type", "application/x-www-form-urlencoded");
                connection.setRequestProperty("User-Agent", Settings.APP_NAME);

                connection.setDoOutput(true);
                connection.setDoInput(true);

                //Read the response

                DataOutputStream os = new DataOutputStream(connection.getOutputStream());
                os.writeBytes(post_data);
                os.close();

                BufferedReader br = null;
                boolean toLog = false;
                if (connection.getResponseCode() >= 400) {
                    httpError = true;
                    br = new BufferedReader(new InputStreamReader((connection.getErrorStream())));
                    toLog = true;
                } else {
                    br = new BufferedReader(new InputStreamReader((connection.getInputStream())));
                }

                String output;

                if (httpError) {
                    LOG.error("Post Data: " + post_data);
                }
                LOG.trace("Query to :" + base + "(method=" + method + ")" + " , HTTP response : \n"); //do not log unless is error > 400
                while ((output = br.readLine()) != null) {
                    LOG.trace(output);
                    answer += output;
                }

                if (httpError) {
                    JSONParser parser = new JSONParser();
                    try {
                        JSONObject obj2 = (JSONObject) (parser.parse(answer));
                        answer = (String) obj2.get("message");

                    } catch (ParseException ex) {
                        LOG.error(ex.toString());
                        LOG.error("answer:" + answer);
                        return null;
                    }
                }
            } //Capture Exceptions
            //2ERROR - Poloniex API returned an error: Nonce must be greater than 14296103443350000. You provided 2. [c.n.n.t.w.PoloniexWrapper:106]

            catch (IllegalStateException ex) {
                LOG.error("IllegalStateException: " + ex.toString());
                return null;
            } catch (NoRouteToHostException | UnknownHostException ex) {
                //Global.BtceExchange.setConnected(false);
                LOG.error("NoRouteToHostException: " + ex.toString());

                answer = TOKEN_BAD_RETURN;
            } catch (IOException ex) {
                LOG.error("IOException: " + ex.toString());
                return null;
            } finally {
                //close the connection, set all objects to null
                connection.disconnect();
            }
            return answer;
        }

    }
}
